/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#if !__SYNTHESIS__
#include <iostream>
#include <stdio.h>
#endif

#include "ap_int.h"
#include "hls_stream.h"
using namespace hls;
#include "hls_interface.h"
#include "hls_utils.h"
#include "hls_ascon.h"

enum fsm_t {S_INIT, S_DATA};

void hls_ciphercore (
    BDO_STREAM& bdo_bus,
    bool *msg_auth_valid,
    BDI_STREAM& bdi_bus,
    KEY_STREAM& key_bus,
    bool decrypt,
    volatile bool key_update
)
{
    /* Top-Level Interface */
#pragma HLS DATA_PACK variable=bdo_bus
#pragma HLS DATA_PACK variable=bdi_bus
#pragma HLS DATA_PACK variable=key_bus
#pragma HLS INTERFACE port=bdo_bus axis
#pragma HLS INTERFACE port=bdi_bus axis
#pragma HLS INTERFACE port=key_bus axis
#pragma HLS INTERFACE port=msg_auth_valid axis

    /* Interface related temporary variables */
    public_bus bdi;
    output_bus bdo;
    secret_bus key;
    array_t bdo_data;

#pragma HLS ARRAY_RESHAPE variable=bdo_data complete dim=1
    array_t bdi_data;
#pragma HLS ARRAY_RESHAPE variable=bdi_data complete dim=1

    /* Static variable */
    static ap_uint<8> key_data[KEYSIZE/8];
#pragma HLS ARRAY_RESHAPE variable=key_data complete dim=1
#if FAIL_RTL
    static ap_uint<8> tag_data[128/8];
#pragma HLS ARRAY_RESHAPE variable=tag_data complete dim=1
#endif

    /* Semi-static variables */
    fsm_t state = S_INIT;

    /* Temporary variables */
    ap_uint<8> S[40];
    array_t to_bdo;
    
    
#pragma HLS ARRAY_RESHAPE variable=S complete dim=1    
#pragma HLS ARRAY_RESHAPE variable=to_bdo complete dim=1    

    ap_uint<6> i;
    bool is_first_dat = 1;
    bool is_after_init = 1;
    bool final_xor;
    bool do_xor_hi, do_xor_lo;
    ap_uint<8> tmp;

    /* Initialization */
    bool loop = 1;
    bdo.size = 0;
#if !__SYNTHESIS__
    size_t dbg_loop = 0;
#endif

    // Key initialization (if needed)
    if (key_update) {
        clear(key_data);
        LOAD_KEY: for (i=0; i<4; i++) {
            key_bus.read(key);
            key_data[i*4+0] = (key.data >> 24) & 0xFF;
            key_data[i*4+1] = (key.data >> 16) & 0xFF;
            key_data[i*4+2] = (key.data >>  8) & 0xFF;
            key_data[i*4+3] = (key.data >>  0) & 0xFF;
        }
    }

    // IV initialization (always)
#if ASCON128a
    S[0] = 0x80;
    S[1] = 0x80;
    S[2] = 0x0c;
    S[3] = 0x08;
#else
    S[0] = 0x80;
    S[1] = 0x40;
    S[2] = 0x0c;
    S[3] = 0x06;
#endif
    for (i=4; i<8; i++) {
    #pragma HLS UNROLL
        S[i] = 0;
    }
#if ASCON128a
    //copy(S+8, key_data);
    clear(S+8);
    bdi_bus.read(bdi);
    reshape_block(bdi_data, bdi.data);
    copy(S+24, bdi_data);
#else
    clear(S+8);
    clear(S+16);
    //  First word
    bdi_bus.read(bdi);
    reshape_block(bdi_data, bdi.data);
    copy(S+24, bdi_data);
    //  Second word
    bdi_bus.read(bdi);
    reshape_block(bdi_data, bdi.data);
    copy(S+32, bdi_data);
#endif

#if DBG
    printState("Init: ", S);
#endif

    MSG_LOOP:
    while (loop) {
#if !__SYNTHESIS__
        // Add this clause to prevent infinite loop during debugging
        if (dbg_loop > 100) return;
        else dbg_loop++;
#endif
        #if DBG
        std::cout << "Current state = " << state << std::endl;
        #endif
        bool is_data = 0;
        bool is_eot = 0;
        ap_uint<4> start = 0;
        
        array_t Srate;
        bool biground;
#pragma HLS ARRAY_RESHAPE variable=Srate complete dim=1    

        // PreProcessing
        if (state == S_INIT) {
#if ASCON128a
            do_xor_hi = 0;
#else
            do_xor_hi = 1;
#endif
            do_xor_lo = 0;
            final_xor = 0;
            copy(Srate, S);
            biground = true;
        } else {
            bdi_bus.read(bdi); reshape_block(bdi_data, bdi.data);
            xor2(to_bdo, S, bdi_data);
	        #if DBG
                printState("\tSin : ", S);
                std::cout << "Reading ... " << std::hex << bdi.data << std::endl;
                std::cout << "\t eoi = " << std::hex << bdi.eoi << std::endl;
                std::cout << "\t eot = " << std::hex << bdi.eot << std::endl;
                //std::cout << "To bdo ";
            #endif

            is_eot = bdi.eot;
            if (bdi.type == TYPE_AD) {
                is_data = 0;
                copy(Srate, to_bdo);
                do_xor_hi = 0;
                do_xor_lo = is_after_init;
                is_after_init = 0;
                final_xor = 0;
                biground = false;
            } else {
                is_data = 1;
                #if DBG
                    std::cout << "Data ... " << std::endl;
                #endif
                if (bdi.size > 0) {
                    bdo.data = reshape_array(to_bdo);
                    bdo_bus.write(bdo);
                    #if DBG
                        std::cout << "Writing ... " << std::hex << bdo.data << std::endl;
                    #endif
                }

                if (decrypt) {
                    for (i=0; i<ELEMENTS; i++) {
                    #pragma HLS UNROLL
                        if ((bdi.valid_bytes >> (ELEMENTS-1-i)) & 0x1) {
                            Srate[i] = bdi_data[i];
                        } else {
                            Srate[i] = to_bdo[i];
                        }
                    }
                } else {
                    copy(Srate, to_bdo);
                }
                do_xor_hi = bdi.eot;
                do_xor_lo = is_after_init;
                final_xor = is_first_dat;
                is_first_dat = 0;
                is_after_init = 0;
                biground = bdi.eot;
            }

        }

        // Assign new S
#if ASCON128a
        for (i=0; i<8; i++) {
        #pragma HLS UNROLL
            S[i] = Srate[i];
            S[i+8] = (state == S_INIT) ? key_data[i] : Srate[i+8];
            ap_uint<8> tmp_hi, tmp_lo;
            if (do_xor_hi) 
                S[16+i] ^= key_data[i];
            else if (state == S_INIT)
                S[16+i] ^= key_data[i+8];
            tmp_hi = (do_xor_hi) ? key_data[8+i] : (ap_uint<8>) 0;
            tmp_lo = (do_xor_lo) ? key_data[  i] : (ap_uint<8>) 0;
            S[24+i] ^= tmp_hi ^ tmp_lo;
        }
        for (i=0; i<7; i++) {
        #pragma HLS UNROLL
            if (do_xor_lo)
                S[32+i] ^= key_data[8+i];
        }
#else
        copy(S, Srate);
        if (do_xor_hi) {
            xor2(S+8,  S+8,  key_data);
            xor2(S+16, S+16, key_data+8);
        }
        if (do_xor_lo) {
            for (i=0; i<15; i++) {
            #pragma HLS UNROLL
                S[24+i] ^= key_data[i];
            }
        }
#endif
        if (do_xor_lo) 
            tmp = key_data[15];
        else 
            tmp = 0;
        S[39] ^= tmp ^ final_xor;

        /* ========= */
        /* MAIN LOOP */
        /* ========= */
        #if DBG
            printState("\tIn  : ", S);
        #endif
        permutation(S, biground);
        #if DBG
            printState("\tOut : ", S);
        #endif

        // Post Processing
        if (state == S_INIT)  {
            state = S_DATA;
        } else {
            if (is_data && is_eot) {
                break;
            }
        }
    }
    bool result = true;

#if ASCON128a
    for (i=0; i<1; i++) {
#else
    for (i=0; i<2; i++) {
#endif
        array_t tag_data;
        xor2(tag_data, S+24+8*i, key_data);
        if (decrypt) {
            #if DBG
                std::cout << "Reading tag ..." << std::endl;
            #endif
            bdi_bus.read(bdi); reshape_block(bdi_data, bdi.data);
            result &= match(bdi_data, tag_data, 8);
        } else {
            #if DBG
                std::cout << "Writing tag ..." << std::endl;
            #endif
            bdo.data = reshape_array(tag_data);
            bdo_bus.write(bdo);
        }
    }
    if (decrypt) *msg_auth_valid = result;

    return;
}
