/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#include "ap_int.h"
#include "hls_interface.h"

#if !__SYNTHESIS__
#include <iostream>
#include <sstream>
#define DBG       1
#define DBG_ROUND 0
#endif

#define ROUND_BIG   12
#if ASCON128a
    #define ROUND_SMALL 8
#else
    #define ROUND_SMALL 6
#endif
#define START_OFFSET ROUND_BIG-ROUND_SMALL
#define ROTR(x,n) (((x)>>(n))|((x)<<(64-(n))))

void printState(std::string text, ap_uint<8>* S) {
#if DBG
    std::cout << std::setfill('0');
    std::cout << text ;
    for (size_t i=0; i<40; i++) {
        std::cout << std::setw(2) << std::hex << std::noshowbase << std::right << (int) S[i];
        if ((i+1)%8 == 0) std::cout << " ";
    }
    std::cout << std::endl;
#endif
}

void printWords(std::string& text, ap_uint<64> x0, ap_uint<64> x1, ap_uint<64> x2, ap_uint<64> x3, ap_uint<64> x4) {
#if DBG
    std::cout << text << ": ";
    std::cout << std::setfill('0');
    std::cout << std::setw(16) << std::hex << std::noshowbase << std::right;
    std::cout << x0 << " "
              << x1 << " "
              << x2 << " "
              << x3 << " "
              << x4 << std::endl;
#endif
}


void load64(ap_uint<64>* x, ap_uint<8>* S) {
#pragma HLS INLINE
  int i;
  *x = 0;
  for (i = 0; i < 8; ++i)
  {
#pragma HLS UNROLL
    *x |= ((ap_uint<64>) S[i]) << (56 - i * 8);
  }
}

void store64(ap_uint<8>* S, ap_uint<64> x) {
#pragma HLS INLINE
  int i;
  for (i = 0; i < 8; ++i)
  {
#pragma HLS UNROLL
    S[i] = (ap_uint<8>) (x >> (56 - i * 8));
  }
}

const ap_uint<8> rc_rom[16] = {
0xf0, 0xe1, 0xd2, 0xc3,
0xb4, 0xa5, 0x96, 0x87,
0x78, 0x69, 0x5a, 0x4b,
0, 0, 0, 0,
};


void oneround(ap_uint<8> *S, int round, bool biground) {
#pragma HLS INLINE
#pragma HLS RESOURCE variable=rc_rom core=ROM_1P_1S
    ap_uint<4> offset = biground ? 0 : START_OFFSET;
    ap_uint<64> x0, x1, x2, x3, x4;
    ap_uint<64> t0, t1, t2, t3, t4;
    load64(&x0, S + 0);
    load64(&x1, S + 8);
    load64(&x2, S + 16);
    load64(&x3, S + 24);
    load64(&x4, S + 32);
#if DBG_ROUND
    std::stringstream sstm;
    sstm << "Rounds - " << std::setw(2) << std::right << i;
    text = sstm.str();
    printWords(text, x0, x1, x2, x3, x4);
#endif
    // addition of round constant
    x2 ^= rc_rom[round + offset];
    // substitution layer
    x0 ^= x4;    x4 ^= x3;    x2 ^= x1;
    t0  = x0;    t1  = x1;    t2  = x2;    t3  = x3;    t4  = x4;
    t0 =~ t0;    t1 =~ t1;    t2 =~ t2;    t3 =~ t3;    t4 =~ t4;
    t0 &= x1;    t1 &= x2;    t2 &= x3;    t3 &= x4;    t4 &= x0;
    x0 ^= t1;    x1 ^= t2;    x2 ^= t3;    x3 ^= t4;    x4 ^= t0;
    x1 ^= x0;    x0 ^= x4;    x3 ^= x2;    x2 =~ x2;
    // linear diffusion layer
    x0 ^= ROTR(x0, 19) ^ ROTR(x0, 28);
    x1 ^= ROTR(x1, 61) ^ ROTR(x1, 39);
    x2 ^= ROTR(x2,  1) ^ ROTR(x2,  6);
    x3 ^= ROTR(x3, 10) ^ ROTR(x3, 17);
    x4 ^= ROTR(x4,  7) ^ ROTR(x4, 41);

    store64(S + 0, x0);
    store64(S + 8, x1);
    store64(S + 16, x2);
    store64(S + 24, x3);
    store64(S + 32, x4);
}

void permutation(ap_uint<8>* S, bool biground) {
#pragma HLS INLINE
#pragma HLS RESOURCE variable=rc_rom core=ROM_1P_1S
  const int rounds = biground ? ROUND_BIG : ROUND_SMALL;

  MAIN_LOOP:
  for (int i = 0; i < rounds; ++i) {
    ap_uint<4> offset = biground ? 0 : START_OFFSET;
    ap_uint<64> x0, x1, x2, x3, x4;
    ap_uint<64> t0, t1, t2, t3, t4;
    load64(&x0, S + 0);
    load64(&x1, S + 8);
    load64(&x2, S + 16);
    load64(&x3, S + 24);
    load64(&x4, S + 32);
#if DBG_ROUND
    std::stringstream sstm;
    sstm << "Rounds - " << std::setw(2) << std::right << i;
    text = sstm.str();
    printWords(text, x0, x1, x2, x3, x4);
#endif
    // addition of round constant
    x2 ^= rc_rom[i + offset];
    // substitution layer
    x0 ^= x4;    x4 ^= x3;    x2 ^= x1;
    t0  = x0;    t1  = x1;    t2  = x2;    t3  = x3;    t4  = x4;
    t0 =~ t0;    t1 =~ t1;    t2 =~ t2;    t3 =~ t3;    t4 =~ t4;
    t0 &= x1;    t1 &= x2;    t2 &= x3;    t3 &= x4;    t4 &= x0;
    x0 ^= t1;    x1 ^= t2;    x2 ^= t3;    x3 ^= t4;    x4 ^= t0;
    x1 ^= x0;    x0 ^= x4;    x3 ^= x2;    x2 =~ x2;
    // linear diffusion layer
    x0 ^= ROTR(x0, 19) ^ ROTR(x0, 28);
    x1 ^= ROTR(x1, 61) ^ ROTR(x1, 39);
    x2 ^= ROTR(x2,  1) ^ ROTR(x2,  6);
    x3 ^= ROTR(x3, 10) ^ ROTR(x3, 17);
    x4 ^= ROTR(x4,  7) ^ ROTR(x4, 41);

    store64(S + 0, x0);
    store64(S + 8, x1);
    store64(S + 16, x2);
    store64(S + 24, x3);
    store64(S + 32, x4);
  }
}

void permutation_v2(ap_uint<8>* S, bool biground) {
#pragma HLS INLINE
  
#if 1
   int rounds = biground ? ROUND_BIG : ROUND_SMALL;
   MAIN_LOOP:
   for (int i = 0; i < rounds; ++i) {
       oneround(S, i, biground);
   }
#else
// #pragma HLS allocation instances=oneround limit=1 function
    oneround(S, 0, biground);
    oneround(S, 1, biground);
    oneround(S, 2, biground);
    oneround(S, 3, biground);
    oneround(S, 4, biground);
    oneround(S, 5, biground);
    oneround(S, 6, biground);
    oneround(S, 7, biground);
    if (biground) {
        oneround(S, 8, biground);
        oneround(S, 9, biground);
        oneround(S, 10, biground);
        oneround(S, 11, biground);
    }
#endif
}

