/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#include "ap_int.h"
#include "hls_utils.h"

#if !__SYNTHESIS__
#include <iostream>
void pp(const char *txt, array_t data)
{
    std::cout << std::hex << txt << reshape_array(data) << std::endl;
}
#endif

void swap_endian(array_t c, array_t a, const int word_bytes)
{
    /* Swab endian within a word */
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=a           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=c           complete dim=1

    size_t tail = ELEMENTS/word_bytes;
    for (size_t i=0; i<tail; i++) {
    #pragma HLS UNROLL
        for (size_t j=0; j<word_bytes; j++) {
        #pragma HLS UNROLL
            c[i*word_bytes + word_bytes-j-1] = a[i*word_bytes+j];
        }
    }
}

void byte_valid(array_t c, array_t a, ap_uint<BLOCKSIZE/8> vld) 
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=a           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=c           complete dim=1

    for (size_t i=0; i<BLOCKSIZE/8; i++) {
    #pragma HLS UNROLL
        if ((vld >> BLOCKSIZE/8-i-1) & 0x1)
            c[i] = a[i];
        else
            c[i] = 0;
    }
}

void byte_valid_sel(array_t c, array_t a, array_t b, ap_uint<BLOCKSIZE/8> vld) 
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=a           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=b           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=c           complete dim=1

    for (size_t i=0; i<BLOCKSIZE/8; i++) {
    #pragma HLS UNROLL
        if ((vld >> BLOCKSIZE/8-i-1) & 0x1)
            c[i] = a[i];
        else
            c[i] = b[i];
    }
}

void byte_valid_pad(
    array_t c, array_t a, 
    ap_uint<BLOCKSIZE/8> vld, 
    ap_uint<BLOCKSIZE/8> pad) 
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=a           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=c           complete dim=1
    ap_uint<8> abyte[BLOCKSIZE/8];
    ap_uint<8> byte[BLOCKSIZE/8];
    data_t tmp = 0;
    int i;

#if DBG
    std::cout << vld << std::endl;
#endif

    for (size_t i=0; i<BLOCKSIZE/8; i++) {
    #pragma HLS UNROLL
        if ((vld >> BLOCKSIZE/8-i-1) & 0x1)
            c[i] = a[i];
        else
            c[i] = ((pad >> BLOCKSIZE/8-i-1) & 0x1) << 7;
    }
}

void clear(array_t a) 
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=a complete dim=1
    int i;
    for(i=0; i<ELEMENTS; i++) {
    #pragma HLS UNROLL
        a[i] = 0;
    }
}

void copy(array_t b, array_t a) 
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=a complete dim=1
#pragma HLS ARRAY_RESHAPE variable=b complete dim=1
    int i;
    for(i=0; i<ELEMENTS; i++) {
    #pragma HLS UNROLL
        b[i] = a[i];
    }
}

#if !__SYNTHESIS__
void display(array_t a)
{
    int i;
    std::cout << std::setfill('0');
    std::cout << std::setw(2) << std::hex << std::noshowbase << std::right;
    for(i=0;i <ELEMENTS; i++) {
        std::cout << a[i];
        if ((i+1)%8 == 0) 
            std::cout << " ";
    }
    std::cout << std::endl;
}
#endif

bool match(array_t a, array_t b, int cnt) 
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=a           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=b           complete dim=1
    int i;
    bool result = 1; 
    for(i=0; i<cnt; i++) {
    #pragma HLS UNROLL
        if (a[i] != b[i]) result = 0; 
    }
    return result;
}


void rot_left(array_t out, array_t in, ap_uint<4> size)
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=in           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=out          complete dim=1
    data_t iblock;
    data_t s1, s2, s3, s4;

    iblock = reshape_array(in);

    if (size & 0x1)
        s1 = (iblock << 8) ^ (iblock >> (128-8));
    else
        s1 = iblock;

    if ((size >> 1) & 0x1)
        s2 = (s1 << 16) ^ (s1 >> (128-16));
    else
        s2 = s1;

    if ((size >> 2) & 0x1)
        s3 = (s2 << 32) ^ (s2 >> (128-32));
    else
        s3 = s2;

    if ((size >> 3) & 0x1)
        s4 = (s3 << 64) ^ (s3 >> (128-64));
    else
        s4 = s3;

    reshape_block(out, s4);
}

void rot_right(array_t out, array_t in, ap_uint<4> size)
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=in           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=out          complete dim=1
    data_t iblock;
    data_t s1, s2, s3, s4;

    iblock = reshape_array(in);

    if (size & 0x1)
        s1 = (iblock >> 8) ^ (iblock << (128-8));
    else
        s1 = iblock;

    if ((size >> 1) & 0x1)
        s2 = (s1 >> 16) ^ (s1 << (128-16));
    else
        s2 = s1;

    if ((size >> 2) & 0x1)
        s3 = (s2 >> 32) ^ (s2 << (128-32));
    else
        s3 = s2;

    if ((size >> 3) & 0x1)
        s4 = (s3 >> 64) ^ (s3 << (128-64));
    else
        s4 = s3;

    reshape_block(out, s4);
}

void shf_left(array_t out, array_t in, ap_uint<4> size)
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=in           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=out          complete dim=1
    data_t iblock;
    data_t s1, s2, s3, s4;

    iblock = reshape_array(in);

    if (size & 0x1)
        s1 = (iblock << 8);
    else
        s1 = iblock;

    if ((size >> 1) & 0x1)
        s2 = (s1 << 16);
    else
        s2 = s1;

    if ((size >> 2) & 0x1)
        s3 = (s2 << 32);
    else
        s3 = s2;

    if ((size >> 3) & 0x1)
        s4 = (s3 << 64);
    else
        s4 = s3;

    reshape_block(out, s4);
}

void shf_left1(array_t out, array_t in)
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=in           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=out          complete dim=1
    data_t iblock, outBlock;
    data_t s1, s2, s3, s4;

    iblock = reshape_array(in);
    outBlock = iblock << 1;

    reshape_block(out, outBlock);
}


void shf_right(array_t out, array_t in, ap_uint<4> size)
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=in           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=out          complete dim=1
    data_t iblock;
    data_t s1, s2, s3, s4;

    iblock = reshape_array(in);

    if (size & 0x1)
        s1 = (iblock >> 8);
    else
        s1 = iblock;

    if ((size >> 1) & 0x1)
        s2 = (s1 >> 16);
    else
        s2 = s1;

    if ((size >> 2) & 0x1)
        s3 = (s2 >> 32);
    else
        s3 = s2;

    if ((size >> 3) & 0x1)
        s4 = (s3 >> 64);
    else
        s4 = s3;

    reshape_block(out, s4);
}

void shf_right1(array_t out, array_t in)
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=in           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=out          complete dim=1
    data_t iblock, outBlock;
    data_t s1, s2, s3, s4;

    iblock = reshape_array(in);
    outBlock = iblock >> 1;

    reshape_block(out, outBlock);
}


void barrel_shf_right1(array_t out, array_t in, ap_uint<7> size)
{
#pragma HLS INLINE
    data_t iblock;
    data_t s1, s2, s3, s4, s5, s6, s7;

    iblock = reshape_array(in);

    if (size & 0x1)
        s1 = (iblock >> 1);
    else
        s1 = iblock;

    if ((size >> 1) & 0x1)
        s2 = (s1 >> 2);
    else
        s2 = s1;

    if ((size >> 2) & 0x1)
        s3 = (s2 >> 4);
    else
        s3 = s2;

    if ((size >> 3) & 0x1)
        s4 = (s3 >> 8);
    else
        s4 = s3;

    if ((size >> 4) & 0x1)
        s5 = (s4 >> 16);
    else
        s5 = s4;

    if ((size >> 5) & 0x1)
        s6 = (s5 >> 32);
    else
        s6 = s5;

    if ((size >> 6) & 0x1)
        s7 = (s6 >> 64);
    else
        s7 = s6;

    reshape_block(out, s7);
}



data_t reshape_array(array_t word)
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=word           complete dim=1
    int i;
    data_t tmp = 0;
    for (i=0; i<ELEMENTS; i++) {
    #pragma HLS UNROLL
        tmp ^= ((data_t) word[i]) << ((ELEMENTS-i-1)*ELEMENTSIZE);
    }
    return tmp;
}

void reshape_block(array_t word, data_t block)
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=word           complete dim=1
    int i;
    for (i=0; i<ELEMENTS; i++) {
    #pragma HLS UNROLL
        word[i] = (ap_uint<ELEMENTSIZE>) (block >> ((ELEMENTS-i-1)*ELEMENTSIZE));
    }
}

void xor2(array_t c, array_t a, array_t b) 
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=a           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=b           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=c           complete dim=1
    int i;
    for(i=0; i<ELEMENTS; i++) {
    #pragma HLS UNROLL
        c[i] = a[i]^b[i];
    }
}

void xor3(array_t d, array_t a, array_t b, array_t c)
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=a           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=b           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=c           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=d           complete dim=1
    int i;
    for(i=0; i<ELEMENTS; i++) {
    #pragma HLS UNROLL
        d[i] = (a[i]^b[i])^c[i];
    }
}


